"""Emit all Myrial/SQL keywords as lowercase strings."""

from raco.myrial.scanner import (builtins, keywords,
                                 types, comprehension_keywords,
                                 word_operators)
from raco.expression.expressions_library import EXPRESSIONS


def get_keywords():
    """Return a list of Myrial/SQL keywords.

    This includes reserved lex tokens and system-defined functions.
    """
    return {
        'builtins': sorted(
            EXPRESSIONS.keys() + [kw.lower() for kw in builtins]),
        'keywords': sorted(kw.lower() for kw in keywords),
        'types': sorted(kw.lower() for kw in types),
        'comprehension_keywords': sorted(
            kw.lower() for kw in comprehension_keywords),
        'word_operators': sorted(kw.lower() for kw in word_operators),
    }



"""
Utility functions for use in Raco expressions
"""

from .expression import (Expression, BinaryOperator, AttributeRef,
                         NamedAttributeRef, UnnamedAttributeRef,
                         NamedStateAttributeRef)
from .aggregate import BuiltinAggregateExpression, AggregateExpression

import copy
import inspect


class NestedAggregateException(Exception):
    """Nested aggregate functions are not allowed"""
    def __init__(self, lineno):
        self.lineno = lineno

    def __str__(self):
        return "Nested aggregate expression on line %d" % self.lineno


def ensure_unnamed(exprs, op):
    """Returns the expressions in the supplied exprs, ensuring that all
    attribute references are in the unnamed perspective. The expressions are
    resolved against the scheme of the supplied operator, but only if
    necessary.

    :param exprs: an expression or sequence of expressions.
    :param op: the operator against whose scheme non-unnamed references will be
               resolved. Scheme is only computed if needed.
    """
    if isinstance(exprs, Expression):
        # exprs is just a single expression
        if only_unnamed_refs(exprs):
            return exprs
        return to_unnamed_recursive(exprs, op.scheme())

    else:
        # exprs is a sequence of expressions.
        if all(only_unnamed_refs(e) for e in exprs):
            return exprs
        op_scheme = op.scheme()
        return [to_unnamed_recursive(e, op_scheme) for e in exprs]


def toUnnamed(ref, scheme):
    """Convert a reference to the unnamed perspective"""
    if issubclass(ref.__class__, UnnamedAttributeRef):
        return ref
    elif issubclass(ref.__class__, NamedAttributeRef):
        return UnnamedAttributeRef(scheme.getPosition(ref.name))
    else:
        raise TypeError("Unknown value reference %s.  Expected a position reference or an attribute reference." % ref)  # noqa


def toNamed(ref, scheme):
    """Convert a reference to the named perspective"""
    if issubclass(ref.__class__, UnnamedAttributeRef):
        attrname = scheme[ref.position][0]
        return NamedAttributeRef(attrname)
    elif issubclass(ref.__class__, NamedAttributeRef):
        return ref
    else:
        raise TypeError("Unknown value reference %s.  Expected a position reference or an attribute reference.")  # noqa


def to_unnamed_recursive(sexpr, scheme):
    """Convert all named column references to unnamed column references."""
    def convert(n):
        if isinstance(n, NamedAttributeRef):
            n = toUnnamed(n, scheme)
        n.apply(convert)
        return n
    return convert(sexpr)


def only_unnamed_refs(sexpr):
    """Returns True if all AttributeRefs in the expression are unnamed."""
    return all(isinstance(e, UnnamedAttributeRef)
               for e in sexpr.walk()
               if isinstance(e, AttributeRef))


def all_classes():
    """Return a list of all classes in the module"""
    import raco.expression as expr
    return [obj for _, obj in inspect.getmembers(expr, inspect.isclass)]


def aggregate_functions():
    """Return all the classes that can be used to construct an aggregate expression"""  # noqa
    allclasses = all_classes()
    opclasses = [opclass for opclass in allclasses
                 if issubclass(opclass, BuiltinAggregateExpression)
                 and not inspect.isabstract(opclass)]

    return opclasses


def binary_ops():
    """Return a list of all classes used to construct arithmetic, like PLUS, DIVIDE, etc."""  # noqa
    allclasses = all_classes()
    opclasses = [opclass for opclass in allclasses
                 if issubclass(opclass, BinaryOperator)
                 and not inspect.isabstract(opclass)]
    return opclasses


def udf_undefined_vars(expr, vars):
    """Return a list of undefined variables in a UDF.

    :param expr: An expression corresponding to a UDF.  Variable references are
    identified by instances of NamedAttributeRef.

    :param vars: A list of variables in the argument list to the function.
    :type vars: list of strings
    """
    return [ex.name for ex in expr.walk()
            if isinstance(ex, NamedAttributeRef) and ex.name not in vars]


def resolve_function(func_expr, arg_dict):
    """Bind variables to arguments in a function invocation.

    :param func_expr: An expression corresponding to function
    :type func_expr: Expression
    :param arg_dict: The arguments to the function
    :type arg_dict: A dictionary mapping string to Expression
    :returns: An expression with no variables
    """

    def convert(n):
        if isinstance(n, NamedAttributeRef):
            n = copy.deepcopy(arg_dict[n.name])
        else:
            n.apply(convert)
        return n

    return convert(copy.deepcopy(func_expr))


def resolve_state_vars(expr, state_vars, mangled_names):
    """Convert references to state variables into NamedStateAttributeRef
    references.

    :param expr: An expression instances
    :type expr: Expression
    :param state_vars: Variables that represent state variables
    :type state_vars: List of strings
    :param mangled_names: A mapping from name to its mangled version
    :type mangled_names: Dictionary from string to string
    :return: An instance of Expression
    """

    def convert(n):
        if isinstance(n, NamedAttributeRef) and n.name in state_vars:
            n = NamedStateAttributeRef(mangled_names[n.name])
        else:
            n.apply(convert)
        return n

    return convert(copy.deepcopy(expr))


def accessed_columns(expr):
    """Return a set of column indexes accessed by an expression.

    Assumes that named attribute references have been converted to integer
    positions.
    """
    for ex in expr.walk():
        assert not isinstance(ex, NamedAttributeRef)

    return set([ex.position for ex in expr.walk()
                if isinstance(ex, UnnamedAttributeRef)])


def rebase_expr(expr, offset):
    """Subtract the given offset from each column access.

    Assumes that named attribute references have been converted to integer
    positions.
    """
    assert offset > 0

    for ex in expr.walk():
        assert not isinstance(ex, NamedAttributeRef)
        if isinstance(ex, UnnamedAttributeRef):
            ex.position -= offset


def reindex_expr(expr, index_map):
    """Changes references to key columns to references to value columns in
    index_map.

    Assumes that named attribute references have been converted to integer
    positions.
    """
    for ex in expr.walk():
        assert (not isinstance(ex, AttributeRef)
                or isinstance(ex, UnnamedAttributeRef))
        if isinstance(ex, UnnamedAttributeRef) and ex.position in index_map:
            ex.position = index_map[ex.position]


def expression_contains_aggregate(ex):
    """Return True if the expression contains an aggregate."""
    return any(isinstance(sx, AggregateExpression) for sx in ex.walk())


def check_no_aggregate(ex, lineno):
    """Raise an exception if the provided expression contains an aggregate."""
    if expression_contains_aggregate(ex):
        raise NestedAggregateException(lineno)


def check_no_nested_aggregate(ex, lineno):
    """Raise an exception if the expression contains a nested aggregate."""

    def descend(sx, in_aggregate):
        is_aggregate = isinstance(sx, AggregateExpression)
        if is_aggregate and in_aggregate:
            raise NestedAggregateException(lineno)
        for child in sx.get_children():
            descend(child, is_aggregate or in_aggregate)

    descend(ex, False)



"""
A RACO language to compile expressions to SQL.
"""

from sqlalchemy import (Column, Table, MetaData, Integer, String,
                        Float, Boolean, DateTime, create_engine, select, func)

import raco.algebra as algebra
from raco.catalog import Catalog
import raco.expression as expression
import raco.scheme as scheme
import raco.types as types


type_to_raco = {Integer: types.LONG_TYPE,
                String: types.STRING_TYPE,
                Float: types.FLOAT_TYPE,
                Boolean: types.BOOLEAN_TYPE,
                DateTime: types.DATETIME_TYPE}


raco_to_type = {types.LONG_TYPE: Integer,
                types.INT_TYPE: Integer,
                types.STRING_TYPE: String,
                types.FLOAT_TYPE: Float,
                types.DOUBLE_TYPE: Float,
                types.BOOLEAN_TYPE: Boolean,
                types.DATETIME_TYPE: DateTime}


class SQLCatalog(Catalog):
    def __init__(self, engine=None):
        self.engine = engine
        self.metadata = MetaData()

    @staticmethod
    def get_num_servers():
        """ Return number of servers in myria deployment """
        return 1

    def num_tuples(self, rel_key):
        """ Return number of tuples of rel_key """
        table = self.metadata.tables[str(rel_key)]
        return self.engine.execute(table.count()).scalar()

    def get_scheme(self, rel_key):
        table = self.metadata.tables[str(rel_key)]
        return scheme.Scheme((c.name, type_to_raco[type(c.type)])
                             for c in table.columns)

    def add_table(self, name, schema):
        columns = [Column(n, raco_to_type[t](), nullable=False)
                   for n, t in schema.attributes]
        # Adds the table to the metadata
        Table(name, self.metadata, *columns)

    def add_tuples(self, name, schema, tuples=None):
        table = self.metadata.tables[name]
        table.create(self.engine)
        if tuples:
            tuples = [{n: v for n, v in zip(schema.get_names(), tup)}
                      for tup in tuples]
            self.engine.execute(table.insert(), tuples)

    def _convert_expr(self, cols, expr, input_scheme):
        if isinstance(expr, expression.AttributeRef):
            return self._convert_attribute_ref(cols, expr, input_scheme)
        if isinstance(expr, expression.ZeroaryOperator):
            return self._convert_zeroary_expr(cols, expr, input_scheme)
        if isinstance(expr, expression.UnaryOperator):
            return self._convert_unary_expr(cols, expr, input_scheme)
        if isinstance(expr, expression.BinaryOperator):
            return self._convert_binary_expr(cols, expr, input_scheme)

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_attribute_ref(self, cols, expr, input_scheme):
        if isinstance(expr, expression.NamedAttributeRef):
            expr = expression.toUnnamed(expr, input_scheme)

        if isinstance(expr, expression.UnnamedAttributeRef):
            # Not an elif since the first may actually turn into a UARef
            return cols[expr.position]

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_zeroary_expr(self, cols, expr, input_scheme):
        if isinstance(expr, expression.COUNTALL):
            return func.count(cols[0])
        if isinstance(expr, expression.Literal):
            return expr.value
        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_unary_expr(self, cols, expr, input_scheme):
        input = self._convert_expr(cols, expr.input, input_scheme)
        if isinstance(expr, expression.MAX):
            return func.max(input)
        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _convert_binary_expr(self, cols, expr, input_scheme):
        left = self._convert_expr(cols, expr.left, input_scheme)
        right = self._convert_expr(cols, expr.right, input_scheme)

        if isinstance(expr, expression.AND):
            return left & right
        if isinstance(expr, expression.OR):
            return left | right
        if isinstance(expr, expression.EQ):
            return left == right
        if isinstance(expr, expression.NEQ):
            return left != right
        if isinstance(expr, expression.LT):
            return left < right
        if isinstance(expr, expression.LTEQ):
            return left <= right
        if isinstance(expr, expression.GT):
            return left > right
        if isinstance(expr, expression.GTEQ):
            return left >= right

        raise NotImplementedError("expression {} to sql".format(type(expr)))

    def _get_zeroary_sql(self, plan):
        if isinstance(plan, algebra.Scan):
            if str(plan.relation_key) not in self.metadata.tables:
                self.add_table(str(plan.relation_key), plan.scheme())
            return self.metadata.tables[str(plan.relation_key)].select()
        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def _get_unary_sql(self, plan):
        input = self.get_sql(plan.input).alias("input")
        input_sch = plan.input.scheme()
        cols = list(input.c)

        if isinstance(plan, algebra.Select):
            cond = self._convert_expr(cols, plan.condition, input_sch)
            return input.select(cond)

        elif isinstance(plan, algebra.Apply):
            clause = [self._convert_expr(cols, e, input_sch).label(name)
                      for (name, e) in plan.emitters]
            return select(clause)

        elif isinstance(plan, algebra.GroupBy):
            a = [self._convert_expr(cols, e, input_sch)
                 for e in plan.aggregate_list]
            g = [self._convert_expr(cols, e, input_sch)
                 for e in plan.grouping_list]
            sel = select(g + a)
            if not plan.grouping_list:
                return sel
            return sel.group_by(*g)

        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def _get_binary_sql(self, plan):
        # Use aliases to resolve duplicate names
        left = self.get_sql(plan.left).alias("left")
        right = self.get_sql(plan.right).alias("right")
        all_cols = left.c + right.c
        all_sch = plan.left.scheme() + plan.right.scheme()
        assert len(all_cols) == len(all_sch)

        if isinstance(plan, algebra.CrossProduct):
            out_cols = zip(all_cols, all_sch.get_names())
            return select([col.label(name) for col, name in out_cols])

        elif isinstance(plan, algebra.ProjectingJoin):
            cond = self._convert_expr(all_cols, plan.condition, all_sch)
            return left.join(right, cond)

        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def _get_nary_sql(self, plan):
        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def get_sql(self, plan):
        if isinstance(plan, algebra.ZeroaryOperator):
            return self._get_zeroary_sql(plan)
        elif isinstance(plan, algebra.UnaryOperator):
            return self._get_unary_sql(plan)
        elif isinstance(plan, algebra.BinaryOperator):
            return self._get_binary_sql(plan)
        elif isinstance(plan, algebra.NaryOperator):
            return self._get_nary_sql(plan)
        raise NotImplementedError("convert {op} to sql".format(op=type(plan)))

    def evaluate(self, plan):
        statement = self.get_sql(plan)
        print statement
        return (tuple(t) for t in self.engine.execute(statement))


import raco.myrial.groupby as groupby
import raco.myrial.multiway as multiway
from raco.myrial.cfg import ControlFlowGraph
from raco.myrial.emitarg import FullWildcardEmitArg, TableWildcardEmitArg
from raco.myrial.exceptions import *
import raco.algebra
import raco.expression
import raco.catalog
import raco.scheme
from raco.language.myrialang import (MyriaLeftDeepTreeAlgebra,
                                     MyriaHyperCubeAlgebra)
from raco.language.myrialang import compile_to_json
from raco.compile import optimize
from raco import relation_key
from raco.expression import StateVar

import collections
import copy


class DuplicateAliasException(Exception):
    """Bag comprehension arguments must have different alias names."""
    pass


class InvalidStatementException(Exception):
    pass


class NoSuchRelationException(Exception):
    pass


def get_unnamed_ref(column_ref, scheme, offset=0):
    """Convert a string or int into an attribute ref on the new table"""  # noqa
    if isinstance(column_ref, int):
        index = column_ref
    else:
        index = scheme.getPosition(column_ref)
    return raco.expression.UnnamedAttributeRef(index + offset)


def check_binop_compatability(op_name, left, right):
    """Check whether the arguments to an operation are compatible."""
    # Todo: check for type compatibility here?
    # https://github.com/uwescience/raco/issues/213
    if len(left.scheme()) != len(right.scheme()):
        raise SchemaMismatchException(op_name)


def check_assignment_compatability(before, after):
    """Check whether multiple assignments are compatible."""
    # TODO: check for exact schema match -- this is blocked by a general
    # cleanup of raco types.
    check_binop_compatability("assignment", before, after)


class ExpressionProcessor(object):
    """Convert syntactic expressions into relational algebra operations."""
    def __init__(self, symbols, catalog, use_dummy_schema=False):
        self.symbols = symbols
        self.catalog = catalog
        self.use_dummy_schema = use_dummy_schema

        # Variables accessed by the current operation
        self.uses_set = set()

    def get_and_clear_uses_set(self):
        """Retrieve the uses set and then clear its value."""
        try:
            return self.uses_set
        finally:
            self.uses_set = set()

    def evaluate(self, expr):
        method = getattr(self, expr[0].lower())
        return method(*expr[1:])

    def __lookup_symbol(self, _id):
        self.uses_set.add(_id)
        return copy.deepcopy(self.symbols[_id])

    def alias(self, _id):
        return self.__lookup_symbol(_id)

    def scan(self, rel_key):
        """Scan a database table."""
        assert isinstance(rel_key, relation_key.RelationKey)
        try:
            scheme = self.catalog.get_scheme(rel_key)
        except KeyError:
            if not self.use_dummy_schema:
                raise NoSuchRelationException(rel_key)

            # Create a dummy schema suitable for emitting plans
            scheme = raco.scheme.DummyScheme()

        return raco.algebra.Scan(rel_key, scheme,
                                 self.catalog.num_tuples(rel_key))

    def load(self, path, scheme):
        return raco.algebra.FileScan(path, scheme)

    def table(self, emit_clause):
        """Emit a single-row table literal."""
        emit_args = []
        for clause in emit_clause:
            emit_args.extend(clause.expand({}))

        from_args = collections.OrderedDict()
        from_args['$$SINGLETON$$'] = raco.algebra.SingletonRelation()

        # Add unbox relations to the from_args dictionary
        for name, sexpr in emit_args:
            self.extract_unbox_args(from_args, sexpr)

        op, info = multiway.merge(from_args)

        # rewrite clauses in terms of the new schema
        emit_args = [(name, multiway.rewrite_refs(sexpr, from_args, info))
                     for (name, sexpr) in emit_args]

        return raco.algebra.Apply(emitters=emit_args, input=op)

    @staticmethod
    def empty(_scheme):
        if not _scheme:
            _scheme = raco.scheme.Scheme()
        return raco.algebra.EmptyRelation(_scheme)

    def select(self, args):
        """Evaluate a select-from-where expression."""
        op = self.bagcomp(args.from_, args.where, args.select)
        if args.distinct:
            op = raco.algebra.Distinct(input=op)
        if args.limit is not None:
            op = raco.algebra.Limit(input=op, count=args.limit)

        return op

    def extract_unbox_args(self, from_args, sexpr):
        for sub_expr in sexpr.walk():
            if isinstance(sub_expr, raco.expression.Unbox):
                rex = sub_expr.relational_expression
                if rex not in from_args:
                    unbox_op = self.evaluate(rex)
                    from_args[rex] = unbox_op

    def bagcomp(self, from_clause, where_clause, emit_clause):
        """Evaluate a bag comprehension.

        from_clause: A list of tuples of the form (id, expr).  expr can
        be None, which means "read the value from the symbol table".

        where_clause: An optional scalar expression (raco.expression).

        emit_clause: A list of EmitArg instances, each defining one or more
        output columns.
        """

        # Make sure no aliases were reused: [FROM X, X EMIT *] is illegal
        from_aliases = set([x[0] for x in from_clause])
        if len(from_aliases) != len(from_clause):
            raise DuplicateAliasException()

        # For each FROM argument, create a mapping from ID to operator
        # (id, raco.algebra.Operator)
        from_args = collections.OrderedDict()

        for _id, expr in from_clause:
            if expr:
                from_args[_id] = self.evaluate(expr)
            else:
                from_args[_id] = self.__lookup_symbol(_id)

        # Expand wildcards into a list of output columns
        assert emit_clause  # There should always be something to emit
        emit_args = []
        statemods = []
        for clause in emit_clause:
            emit_args.extend(clause.expand(from_args))
            statemods.extend(clause.get_statemods())

        orig_op, _info = multiway.merge(from_args)
        orig_schema_length = len(orig_op.scheme())

        # Add unbox relations to the from_args dictionary
        for name, sexpr in emit_args:
            self.extract_unbox_args(from_args, sexpr)
        if where_clause:
            self.extract_unbox_args(from_args, where_clause)

        # Create a single RA operation that is the cross of all targets
        op, info = multiway.merge(from_args)

        # HACK: calculate unboxed columns as implicit grouping columns,
        # so they can be used in grouping terms.
        new_schema_length = len(op.scheme())
        implicit_group_by_cols = range(orig_schema_length, new_schema_length)

        ################################################
        # Compile away unbox expressions in where, emit clauses
        ################################################

        if where_clause:
            where_clause = multiway.rewrite_refs(where_clause, from_args, info)
            # Extract the type of there where clause to force type safety
            # to be checked
            where_clause.typeof(op.scheme(), None)
            op = raco.algebra.Select(condition=where_clause, input=op)

        emit_args = [(name, multiway.rewrite_refs(sexpr, from_args, info))
                     for (name, sexpr) in emit_args]

        statemods = multiway.rewrite_statemods(statemods, from_args, info)

        if any(raco.expression.expression_contains_aggregate(ex)
               for name, ex in emit_args):
            return groupby.groupby(op, emit_args, implicit_group_by_cols,
                                   statemods)
        else:
            if statemods:
                return raco.algebra.StatefulApply(emit_args, statemods, op)
            if (len(from_args) == 1 and len(emit_clause) == 1 and
                isinstance(emit_clause[0],
                           (TableWildcardEmitArg, FullWildcardEmitArg))):
                return op
            return raco.algebra.Apply(emit_args, op)

    def distinct(self, expr):
        op = self.evaluate(expr)
        return raco.algebra.Distinct(input=op)

    def unionall(self, e1, e2):
        left = self.evaluate(e1)
        right = self.evaluate(e2)
        check_binop_compatability("unionall", left, right)
        return raco.algebra.UnionAll(left, right)

    def countall(self, expr):
        op = self.evaluate(expr)
        grouping_list = []
        agg_list = [raco.expression.COUNTALL()]
        return raco.algebra.GroupBy(grouping_list, agg_list, op)

    def intersect(self, e1, e2):
        left = self.evaluate(e1)
        right = self.evaluate(e2)
        check_binop_compatability("intersect", left, right)
        return raco.algebra.Intersection(left, right)

    def diff(self, e1, e2):
        left = self.evaluate(e1)
        right = self.evaluate(e2)
        check_binop_compatability("diff", left, right)
        return raco.algebra.Difference(left, right)

    def limit(self, expr, count):
        op = self.evaluate(expr)
        return raco.algebra.Limit(input=op, count=count)

    def cross(self, left_target, right_target):
        left = self.evaluate(left_target)
        right = self.evaluate(right_target)

        return raco.algebra.CrossProduct(left, right)

    def join(self, left_target, right_target):
        """Convert parser.JoinTarget arguments into a Join operation"""

        left = self.evaluate(left_target.expr)
        right = self.evaluate(right_target.expr)

        assert len(left_target.columns) == len(right_target.columns)

        left_scheme = left.scheme()
        left_refs = [get_unnamed_ref(c, left_scheme, 0)
                     for c in left_target.columns]

        right_scheme = right.scheme()
        right_refs = [get_unnamed_ref(c, right_scheme, len(left_scheme))
                      for c in right_target.columns]

        join_conditions = [raco.expression.EQ(x, y) for x, y in
                           zip(left_refs, right_refs)]

        # Merge the join conditions into a big AND expression

        def andify(x, y):
            """Merge two scalar expressions with an AND"""
            return raco.expression.AND(x, y)

        condition = reduce(andify, join_conditions)
        return raco.algebra.Join(condition, left, right)


class StatementProcessor(object):
    """Evaluate a list of statements"""

    def __init__(self, catalog, use_dummy_schema=False):
        # Map from identifiers (aliases) to raco.algebra.Operation instances
        self.symbols = {}

        assert isinstance(catalog, raco.catalog.Catalog)
        self.catalog = catalog
        self.ep = ExpressionProcessor(self.symbols, catalog, use_dummy_schema)

        self.cfg = ControlFlowGraph()

    def evaluate(self, statements):
        """Evaluate a list of statements"""
        for statement in statements:
            # Switch on the first tuple entry
            method = getattr(self, statement[0].lower())
            method(*statement[1:])

    def __evaluate_expr(self, expr, _def):
        """Evaluate an expression; add a node to the control flow graph.

        :param expr: An expression to evaluate
        :type expr: Myrial AST tuple
        :param _def: The variable defined by the expression, or None for
                     non-statements
        :type _def: string
        """

        op = self.ep.evaluate(expr)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, _def, uses_set)
        return op

    def __do_assignment(self, _id, expr):
        """Process an assignment statement; add a node to the control flow
        graph.

        :param _id: The target variable name.
        :type _id: string
        :param expr: The relational expression to evaluate
        :type expr: A Myrial expression AST node tuple
        """

        child_op = self.ep.evaluate(expr)
        if _id in self.symbols:
            check_assignment_compatability(child_op, self.symbols[_id])

        op = raco.algebra.StoreTemp(_id, child_op)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, _id, uses_set)

        # Point future references of this symbol to a scan of the materialized
        # table. Note that this assumes there is no scoping in Myrial.
        self.symbols[_id] = raco.algebra.ScanTemp(_id, child_op.scheme())

    def assign(self, _id, expr):
        """Map a variable to the value of an expression."""
        self.__do_assignment(_id, expr)

    def store(self, _id, rel_key, how_partitioned):
        assert isinstance(rel_key, relation_key.RelationKey)

        alias_expr = ("ALIAS", _id)
        child_op = self.ep.evaluate(alias_expr)

        if how_partitioned:
            scheme = child_op.scheme()
            col_list = [get_unnamed_ref(a, scheme) for a in how_partitioned]
            child_op = raco.algebra.Shuffle(child_op, col_list)
        op = raco.algebra.Store(rel_key, child_op)

        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, None, uses_set)

    def dump(self, _id):
        alias_expr = ("ALIAS", _id)
        child_op = self.ep.evaluate(alias_expr)
        op = raco.algebra.Dump(child_op)
        uses_set = self.ep.get_and_clear_uses_set()
        self.cfg.add_op(op, None, uses_set)

    def dowhile(self, statement_list, termination_ex):
        first_op_id = self.cfg.next_op_id  # op ID of the top of the loop

        for _type, _id, expr in statement_list:
            if _type != 'ASSIGN':
                # TODO: Better error message
                raise InvalidStatementException('%s not allowed in do/while' %
                                                _type.lower())
            self.__do_assignment(_id, expr)

        last_op_id = self.cfg.next_op_id

        self.__evaluate_expr(termination_ex, None)

        # Add a control flow edge from the loop condition to the top of the
        # loop
        self.cfg.add_edge(last_op_id, first_op_id)

    def get_logical_plan(self, **kwargs):
        """Return an operator representing the logical query plan."""
        return self.cfg.get_logical_plan(
            dead_code_elimination=kwargs.get('dead_code_elimination', True),
            apply_chaining=kwargs.get('apply_chaining', True))

    def __get_physical_plan_for__(self, target_phys_algebra, **kwargs):
        logical_plan = self.get_logical_plan(**kwargs)

        kwargs['target'] = target_phys_algebra
        return optimize(logical_plan, **kwargs)

    def get_physical_plan(self, **kwargs):
        """Return an operator representing the physical query plan."""
        target_phys_algebra = kwargs.get('target_alg')
        if target_phys_algebra is None:
            if kwargs.get('multiway_join', False):
                target_phys_algebra = MyriaHyperCubeAlgebra(self.catalog)
            else:
                target_phys_algebra = MyriaLeftDeepTreeAlgebra()

        return self.__get_physical_plan_for__(target_phys_algebra, **kwargs)

    def get_json(self, **kwargs):
        lp = self.get_logical_plan()
        pps = self.get_physical_plan(**kwargs)
        # TODO This is not correct. The first argument is the raw query string,
        # not the string representation of the logical plan
        return compile_to_json(str(lp), pps, pps, "myrial")



from raco import expression
import raco.types

from collections import OrderedDict


class DummyScheme(object):
    """Dummy scheme used to generate plans in the absence of catalog info."""
    def __len__(self):
        return 0

    def __repr__(self):
        return "DummyScheme()"


class Scheme(object):
    """Add an attribute to the scheme."""
    salt = "1"

    def __init__(self, attributes=None):
        if attributes is None:
            attributes = []
        self.attributes = []
        self.asdict = OrderedDict()
        for n, t in attributes:
            self.addAttribute(n, t)

    def addAttribute(self, name, _type):
        assert _type in raco.types.ALL_TYPES, \
            'Invalid type name: %s' % str(_type)
        _type = raco.types.map_type(_type)

        if name in self.asdict:
            # ugly.  I don't like throwing errors in this case, but it's worse
            # not to
            return self.addAttribute(name + self.salt, _type)
        self.asdict[name] = (len(self.attributes), _type)
        self.attributes.append((name, _type))
        # just in case we changed the name.  ugly.
        return name

    def get_types(self):
        """Return a list of the types in this scheme."""
        return [_type for name, _type in self.attributes]

    def get_names(self):
        """Return a list of the names in this scheme."""
        return [name for name, _type in self.attributes]

    def typecheck(self, tup):
        rmap = raco.types.reverse_python_type_map
        try:
            return all([rmap[_type](v) for (_, _type), v in
                       zip(self.attributes, tup)])
        except:
            raise TypeError("%s not of type %s" % (tup, self.attributes))

    def __eq__(self, other):
        return self.attributes == other.attributes

    def __ne__(self, other):
        return not (self == other)

    def getPosition(self, name):
        return self.asdict[name][0]

    def getName(self, position):
        return self[position][0]

    def getType(self, name):
        if type(name) == int:
            return self[name][1]
        else:
            return self.asdict[name][1]

    def resolve(self, attrref):
        """return the name and type of the attribute reference, resolved
        against this scheme"""
        unnamed = expression.toUnnamed(attrref, self)
        return self.getName(unnamed.position), self.getType(unnamed.position)

    def ascolumnlist(self):
        """Return a columnlist structure suitable for use with Project and
        ProjectingJoin. Currently a list of positional attribute references.
        May eventually be a scheme itself."""
        return [expression.UnnamedAttributeRef(i) for i in xrange(len(self))]

    def __contains__(self, attr, typ=None):
        if typ:
            return (attr, type) in self.attributes
        else:
            return attr in self.asdict

    def __str__(self):
        """Pretty print the scheme"""
        return str(self.attributes)

    def __repr__(self):
        return "Scheme({att!r})".format(att=self.attributes)

    def __len__(self):
        """Return the number of attributes in the scheme"""
        return len(self.attributes)

    def __iter__(self):
        return self.attributes.__iter__()

    def __getitem__(self, key):
        return self.attributes.__getitem__(key)

    def __add__(self, other):
        newsch = Scheme(self.attributes)
        for (n, t) in other:
            newsch.addAttribute(n, t)
        return newsch


class EmptyScheme(Scheme):
    def __init__(self):
        Scheme.__init__(self, [])



from nose.plugins.skip import SkipTest
import subprocess
import sys
import unittest


def check_output_and_print_stderr(args):
    """Run the specified command. If it does not exit cleanly, print the stderr
    of the command to stdout. Note that stderr prints are displayed as tests
    run, whereas stdout prints show up next to the failed test. We want the
    latter."""
    try:
        subprocess.check_output(args, stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError as e:
        print e.output
        raise


class StyleTest(unittest.TestCase):
    """run flake8 with the right arguments and ensure all files pass"""

    def test_style(self):
        "run flake8 with the right arguments and ensure all files pass"
        check_output_and_print_stderr(['flake8', '--ignore=F', 'raco'])

    def test_pylint(self):
        "run pylint -E to catch obvious errors"
        # TODO fix this.
        raise SkipTest()
        check_output_and_print_stderr(['pylint', '-E', 'raco'])




# TODO: To be refactored into parallel shared memory lang,
# where you plugin in the parallel shared memory language specific codegen

from raco import algebra
from raco import expression
from raco.language import Algebra
from raco import rules
from raco.pipelines import Pipelined
from raco.language.clangcommon import StagedTupleRef, ct, CBaseLanguage
from raco.language import clangcommon
from raco.utility import emitlist

from raco.algebra import gensym

import logging
_LOG = logging.getLogger(__name__)

import os.path
import itertools

template_path = os.path.join(os.path.dirname(os.path.abspath(__file__)),
                             "grappa_templates")


def readtemplate(fname):
    return file(os.path.join(template_path, fname)).read()


base_template = readtemplate("base_query.template")


class GrappaStagedTupleRef(StagedTupleRef):
    def __afterDefinitionCode__(self):
        # Grappa requires structures to be block aligned if they will be
        # iterated over with localizing forall
        return "GRAPPA_BLOCK_ALIGNED"


class GrappaLanguage(CBaseLanguage):
    @staticmethod
    def base_template():
        return base_template

    @staticmethod
    def log(txt):
        return """LOG(INFO) << "%s";\n""" % txt

    @staticmethod
    def log_unquoted(code, level=0):
        if level == 0:
            log_str = "LOG(INFO)"
        else:
            log_str = "VLOG(%s)" % (level)

        return """%(log_str)s << %(code)s;\n""" % locals()

    @staticmethod
    def group_wrap(ident, grpcode, attrs):
        pipeline_template = ct("""
        Grappa::Metrics::reset();
        auto start_%(ident)s = walltime();
        %(grpcode)s
        auto end_%(ident)s = walltime();
        auto runtime_%(ident)s = end_%(ident)s - start_%(ident)s;
        %(timer_metric)s += runtime_%(ident)s;
        VLOG(1) << "pipeline group %(ident)s: " << runtime_%(ident)s << " s";
        """)

        timer_metric = None
        if attrs['type'] == 'in_memory':
            timer_metric = "in_memory_runtime"
        elif attrs['type'] == 'scan':
            timer_metric = "saved_scan_runtime"

        code = pipeline_template % locals()
        return code

    @staticmethod
    def pipeline_wrap(ident, plcode, attrs):
        code = plcode

        # timing code
        if True:
            inner_code = code
            timing_template = ct("""auto start_%(ident)s = walltime();
            %(inner_code)s
            auto end_%(ident)s = walltime();
            auto runtime_%(ident)s = end_%(ident)s - start_%(ident)s;
            VLOG(1) << "pipeline %(ident)s: " << runtime_%(ident)s << " s";
            VLOG(1) << "timestamp %(ident)s start " << std::setprecision(15)\
             << start_%(ident)s;
            VLOG(1) << "timestamp %(ident)s end " << std::setprecision(15)\
             << end_%(ident)s;
            """)
            code = timing_template % locals()

        dependences = attrs.get('dependences', set())
        assert isinstance(dependences, set)

        _LOG.debug("pipeline %s dependences %s", ident, dependences)
        dependence_code = emitlist([wait_statement(d) for d in dependences])
        dependence_captures = emitlist(
            [",&{dep}".format(dep=d) for d in dependences])

        code = """{dependence_code}
                  {inner_code}
                  """.format(dependence_code=dependence_code,
                             inner_code=code)

        syncname = attrs.get('sync')
        if syncname:
            inner_code = code
            sync_template = ct("""
            CompletionEvent %(syncname)s;
            spawn(&%(syncname)s, [=%(dependence_captures)s] {
                    %(inner_code)s
                    });
                    """)
            code = sync_template % locals()

        return code

    @classmethod
    def compile_stringliteral(cls, st):
        sid = cls.newstringident()
        decl = """int64_t %s;""" % (sid)
        lookup_init = """auto l_%(sid)s = string_index.string_lookup(%(st)s);
                   on_all_cores([=] { %(sid)s = l_%(sid)s; });""" % locals()
        build_init = """
        string_index = build_string_index("sp2bench_1m.index.medium");
        """

        return """(%s)""" % sid, [decl], [build_init, lookup_init]
        # raise ValueError("String Literals not supported in
        # C language: %s" % s)


class GrappaOperator (Pipelined, algebra.Operator):
    _language = GrappaLanguage

    @classmethod
    def new_tuple_ref(cls, sym, scheme):
        return GrappaStagedTupleRef(sym, scheme)

    @classmethod
    def language(cls):
        return cls._language

    def postorder_traversal(self, func):
        return self.postorder(func)


from raco.algebra import UnaryOperator


def create_pipeline_synchronization(state):
    """
    The pipeline_synchronization will sync tasks
    within a single pipeline. Adds this new object to
    the compiler state.
    """
    global_syncname = gensym()

    # true = tracked by gce user metrics
    global_sync_decl_template = ct("""
        GlobalCompletionEvent %(global_syncname)s(true);
        """)
    global_sync_decl = global_sync_decl_template % locals()

    gce_metric_template = """
    GRAPPA_DEFINE_METRIC(CallbackMetric<int64_t>, \
    app_%(pipeline_id)s_gce_incomplete, []{
    return %(global_syncname)s.incomplete();
    });
    """
    pipeline_id = state.getCurrentPipelineId()
    gce_metric_def = gce_metric_template % locals()

    state.addDeclarations([global_sync_decl, gce_metric_def])

    state.setPipelineProperty('global_syncname', global_syncname)
    return global_syncname


# TODO: replace with ScanTemp functionality?
class GrappaMemoryScan(algebra.UnaryOperator, GrappaOperator):
    def num_tuples(self):
        return 10000  # placeholder

    def produce(self, state):
        self.input.produce(state)

    # TODO: when have pipeline tree representation,
    # will have a consumeMaterialized() method instead;
    # for now we reuse the tuple-based consume
    def consume(self, inputsym, src, state):
        # generate the materialization from file into memory

        # scan from index
        # memory_scan_template = """forall_localized( %(inputsym)s_index->vs, \
        # %(inputsym)s_index->nv, [](int64_t ai, Vertex& a) {
        #      forall_here_async<&impl::local_gce>( 0, a.nadj, \
        # [=](int64_t start, int64_t iters) {
        #      for (int64_t i=start; i<start+iters; i++) {
        #        auto %(tuple_name)s = a.local_adj[i];
        #
        #          %(inner_plan_compiled)s
        #       } // end scan over %(inputsym)s (for)
        #       }); // end scan over %(inputsym)s (forall_here_async)
        #       }); // end scan over %(inputsym)s (forall_localized)
        #       """

        global_syncname = create_pipeline_synchronization(state)
        get_pipeline_task_name(state)

        memory_scan_template = ct("""
    forall<&%(global_syncname)s>( %(inputsym)s.data, %(inputsym)s.numtuples, \
    [=](int64_t i, %(tuple_type)s& %(tuple_name)s) {
    %(inner_plan_compiled)s
    }); // end  scan over %(inputsym)s
    """)

        stagedTuple = state.lookupTupleDef(inputsym)
        tuple_type = stagedTuple.getTupleTypename()
        tuple_name = stagedTuple.name

        inner_plan_compiled = self.parent().consume(stagedTuple, self, state)

        code = memory_scan_template % locals()
        state.setPipelineProperty('type', 'in_memory')
        state.setPipelineProperty('source', self.__class__)
        state.addPipeline(code)
        return None

    def shortStr(self):
        return "%s" % (self.opname())

    def __eq__(self, other):
        """
        See important __eq__ notes below
        @see FileScan.__eq__
        """
        return UnaryOperator.__eq__(self, other)


class GrappaSymmetricHashJoin(algebra.Join, GrappaOperator):
    _i = 0

    @classmethod
    def __genBaseName__(cls):
        name = "%03d" % cls._i
        cls._i += 1
        return name

    def __getHashName__(self):
        name = "dhash_%s" % self.symBase
        return name

    def produce(self, state):
        self.symBase = self.__genBaseName__()

        if not isinstance(self.condition, expression.EQ):
            msg = "The C compiler can only handle equi-join conditions\
             of a single attribute: %s" % self.condition
            raise ValueError(msg)

        init_template = ct("""%(hashname)s.init_global_DHT( &%(hashname)s, \
        cores()*16*1024 );
                        """)
        declr_template = ct("""typedef DoubleDHT<int64_t, \
                                                   %(left_in_tuple_type)s, \
                                                   %(right_in_tuple_type)s,
                                                std_hash> \
                    DHT_%(left_in_tuple_type)s_%(right_in_tuple_type)s;
      DHT_%(left_in_tuple_type)s_%(right_in_tuple_type)s %(hashname)s;
      """)

        my_sch = self.scheme()
        left_sch = self.left.scheme()

        # declaration of hash map
        self._hashname = self.__getHashName__()
        hashname = self._hashname
        self.leftTypeRef = state.createUnresolvedSymbol()
        left_in_tuple_type = self.leftTypeRef.getPlaceholder()
        self.rightTypeRef = state.createUnresolvedSymbol()
        right_in_tuple_type = self.rightTypeRef.getPlaceholder()
        hashdeclr = declr_template % locals()

        state.addDeclarationsUnresolved([hashdeclr])

        self.outTuple = GrappaStagedTupleRef(gensym(), my_sch)
        out_tuple_type_def = self.outTuple.generateDefinition()
        state.addDeclarations([out_tuple_type_def])

        # find the attribute that corresponds to the right child
        self.rightCondIsRightAttr = \
            self.condition.right.position >= len(left_sch)
        self.leftCondIsRightAttr = \
            self.condition.left.position >= len(left_sch)
        assert self.rightCondIsRightAttr ^ self.leftCondIsRightAttr

        self.right.childtag = "right"
        state.addInitializers([init_template % locals()])
        self.right.produce(state)

        self.left.childtag = "left"
        self.left.produce(state)

    def consume(self, t, src, state):
        access_template = ct("""
        %(hashname)s.insert_lookup_iter_%(side)s<&%(global_syncname)s>(\
        %(keyname)s.get(%(keypos)s), %(keyname)s, \
        [=](%(other_tuple_type)s %(valname)s) {
            join_coarse_result_count++;
            %(out_tuple_type)s %(out_tuple_name)s = \
                             combine<%(out_tuple_type)s, \
                                      %(left_type)s, \
                                      %(right_type)s> (%(left_name)s, \
                            %(right_name)s);
                                %(inner_plan_compiled)s
                                });
                                """)

        hashname = self._hashname
        keyname = t.name
        side = src.childtag

        outTuple = self.outTuple
        out_tuple_type = self.outTuple.getTupleTypename()
        out_tuple_name = self.outTuple.name

        global_syncname = state.getPipelineProperty('global_syncname')

        if src.childtag == "right":
            left_sch = self.left.scheme()

            # save for later
            self.right_in_tuple_type = t.getTupleTypename()
            state.resolveSymbol(self.rightTypeRef, self.right_in_tuple_type)

            if self.rightCondIsRightAttr:
                keypos = self.condition.right.position \
                    - len(left_sch)
            else:
                keypos = self.condition.left.position \
                    - len(left_sch)

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            other_tuple_type = self.leftTypeRef.getPlaceholder()
            left_type = other_tuple_type
            right_type = self.right_in_tuple_type
            left_name = gensym()
            right_name = keyname
            self.right_name = right_name
            valname = left_name

            code = access_template % locals()
            return code

        if src.childtag == "left":
            right_in_tuple_type = self.right_in_tuple_type
            left_in_tuple_type = t.getTupleTypename()
            state.resolveSymbol(self.leftTypeRef, left_in_tuple_type)

            if self.rightCondIsRightAttr:
                keypos = self.condition.left.position
            else:
                keypos = self.condition.right.position

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            left_type = left_in_tuple_type
            right_type = self.right_in_tuple_type
            other_tuple_type = self.right_in_tuple_type
            left_name = keyname
            right_name = gensym()
            valname = right_name

            code = access_template % locals()
            return code

        assert False, "src not equal to left or right"


class GrappaShuffleHashJoin(algebra.Join, GrappaOperator):
    _i = 0

    @classmethod
    def __genBaseName__(cls):
        name = "%03d" % cls._i
        cls._i += 1
        return name

    def __getHashName__(self):
        name = "hashjoin_reducer_%s" % self.symBase
        return name

    def produce(self, state):
        left_sch = self.left.scheme()

        self.syncnames = []
        self.symBase = self.__genBaseName__()

        self.right.childtag = "right"
        self.rightTupleTypeRef = None  # may remain None if CSE succeeds
        self.leftTupleTypeRef = None  # may remain None if CSE succeeds

        # find the attribute that corresponds to the right child
        self.rightCondIsRightAttr = \
            self.condition.right.position >= len(left_sch)
        self.leftCondIsRightAttr = \
            self.condition.left.position >= len(left_sch)
        assert self.rightCondIsRightAttr ^ self.leftCondIsRightAttr

        # find right key position
        if self.rightCondIsRightAttr:
            self.right_keypos = self.condition.right.position \
                - len(left_sch)
        else:
            self.right_keypos = self.condition.left.position \
                - len(left_sch)

        # find left key position
        if self.rightCondIsRightAttr:
            self.left_keypos = self.condition.left.position
        else:
            self.left_keypos = self.condition.right.position

        # define output tuple
        outTuple = GrappaStagedTupleRef(gensym(), self.scheme())
        out_tuple_type_def = outTuple.generateDefinition()
        out_tuple_type = outTuple.getTupleTypename()
        out_tuple_name = outTuple.name

        # common index is defined by same right side and same key
        # TODO: probably want also left side
        hashtableInfo = state.lookupExpr((self.right, self.right_keypos))
        if not hashtableInfo:
            # if right child never bound then store hashtable symbol and
            # call right child produce
            self._hashname = self.__getHashName__()
            _LOG.debug("generate hashname %s for %s", self._hashname, self)

            hashname = self._hashname

            # declaration of hash map
            self.rightTupleTypeRef = state.createUnresolvedSymbol()
            self.leftTupleTypeRef = state.createUnresolvedSymbol()
            self.outTupleTypeRef = state.createUnresolvedSymbol()
            right_type = self.rightTupleTypeRef.getPlaceholder()
            left_type = self.leftTupleTypeRef.getPlaceholder()

            # TODO: really want this addInitializers to be addPreCode
            # TODO: *for all pipelines that use this hashname*
            init_template = ct("""
            auto %(hashname)s_num_reducers = cores();
            auto %(hashname)s = allocateJoinReducers\
            <int64_t,%(left_type)s,%(right_type)s,%(out_tuple_type)s>
                (%(hashname)s_num_reducers);
            auto %(hashname)s_ctx = HashJoinContext<int64_t,%(left_type)s,
                %(right_type)s,%(out_tuple_type)s>
                (%(hashname)s, %(hashname)s_num_reducers);""")

            state.addInitializers([init_template % locals()])
            self.right.produce(state)

            self.left.childtag = "left"
            self.left.produce(state)

            state.saveExpr((self.right, self.right_keypos),
                           (self._hashname, right_type, left_type,
                            self.right_syncname, self.left_syncname))

        else:
            # if found a common subexpression on right child then
            # use the same hashtable
            self._hashname, right_type, left_type,\
                self.right_syncname, self.left_syncname = hashtableInfo
            _LOG.debug("reuse hash %s for %s", self._hashname, self)

        # now that Relation is produced, produce its contents by iterating over
        # the join result
        iterate_template = ct("""MapReduce::forall_symmetric
        <&%(pipeline_sync)s>
        (%(hashname)s, &JoinReducer<int64_t,%(left_type)s,
        %(right_type)s,%(out_tuple_type)s>::resultAccessor,
            [=](%(out_tuple_type)s& %(out_tuple_name)s) {
                 %(inner_code_compiled)s
            });
        """)

        hashname = self._hashname

        state.addDeclarations([out_tuple_type_def])

        pipeline_sync = create_pipeline_synchronization(state)
        get_pipeline_task_name(state)

        # add dependences on left and right inputs
        state.addToPipelinePropertySet('dependences', self.right_syncname)
        state.addToPipelinePropertySet('dependences', self.left_syncname)

        # reduce is a single self contained pipeline.
        # future hashjoin implementations may pipeline out of it
        # by passing a continuation to reduceExecute
        reduce_template = ct("""
        %(hashname)s_ctx.reduceExecute();

        """)
        state.addPreCode(reduce_template % locals())

        delete_template = ct("""
            freeJoinReducers(%(hashname)s, %(hashname)s_num_reducers);""")
        state.addPostCode(delete_template % locals())

        inner_code_compiled = self.parent().consume(outTuple, self, state)

        code = iterate_template % locals()
        state.setPipelineProperty('type', 'in_memory')
        state.setPipelineProperty('source', self.__class__)
        state.addPipeline(code)

    def consume(self, inputTuple, fromOp, state):
        if fromOp.childtag == "right":
            side = "Right"
            self.right_syncname = get_pipeline_task_name(state)

            keypos = self.right_keypos

            self.rightTupleTypename = inputTuple.getTupleTypename()
            if self.rightTupleTypeRef is not None:
                state.resolveSymbol(self.rightTupleTypeRef,
                                    self.rightTupleTypename)
        elif fromOp.childtag == "left":
            side = "Left"
            self.left_syncname = get_pipeline_task_name(state)

            keypos = self.left_keypos

            self.leftTupleTypename = inputTuple.getTupleTypename()
            if self.leftTupleTypeRef is not None:
                state.resolveSymbol(self.leftTupleTypeRef,
                                    self.leftTupleTypename)
        else:
            assert False, "src not equal to left or right"

        hashname = self._hashname
        keyname = inputTuple.name
        keytype = inputTuple.getTupleTypename()

        # intra-pipeline sync
        global_syncname = state.getPipelineProperty('global_syncname')

        mat_template = ct("""%(hashname)s_ctx.emitIntermediate%(side)s\
                <&%(global_syncname)s>(\
                %(keyname)s.get(%(keypos)s), %(keyname)s);""")

        # materialization point
        code = mat_template % locals()
        return code


class GrappaGroupBy(algebra.GroupBy, GrappaOperator):
    _i = 0

    @classmethod
    def __genHashName__(cls):
        name = "group_hash_%03d" % cls._i
        cls._i += 1
        return name

    def produce(self, state):
        assert len(self.grouping_list) <= 2, \
            """%s does not currently support \
            "groupings of more than 2 attributes"""\
            % self.__class__.__name__
        assert len(self.aggregate_list) == 1, \
            "%s currently only supports aggregates of 1 attribute"\
            % self.__class__.__name__
        for agg_term in self.aggregate_list:
            assert isinstance(agg_term,
                              expression.BuiltinAggregateExpression), \
                """%s only supports simple aggregate expressions.
                A rule should create Apply[GroupBy]""" \
                % self.__class__.__name__

        self.useKey = len(self.grouping_list) > 0
        _LOG.debug("groupby uses keys? %s" % self.useKey)

        declr_template = None
        if self.useKey:
            if len(self.grouping_list) == 1:
                declr_template = ct("""typedef DHT_symmetric<int64_t, \
                                  int64_t, std_hash> \
                                   DHT_int64;
                """)
            elif len(self.grouping_list) == 2:
                declr_template = ct("""typedef DHT_symmetric<\
                std::pair<int64_t,int64_t>, \
                                  int64_t, pair_hash> \
                                   DHT_pair_int64;
                """)

        self._hashname = self.__genHashName__()
        _LOG.debug("generate hashname %s for %s", self._hashname, self)

        hashname = self._hashname

        if declr_template is not None:
            hashdeclr = declr_template % locals()
            state.addDeclarationsUnresolved([hashdeclr])

        if self.useKey:
            if len(self.grouping_list) == 1:
                init_template = ct("""auto %(hashname)s = \
                DHT_int64::create_DHT_symmetric( );""")
            elif len(self.grouping_list) == 2:
                init_template = ct("""auto %(hashname)s = \
                DHT_pair_int64::create_DHT_symmetric( );""")

        else:
            init_template = ct("""auto %(hashname)s = counter::create();
            """)

        state.addInitializers([init_template % locals()])

        self.input.produce(state)

        # now that everything is aggregated, produce the tuples
        assert len(self.column_list()) == 1 \
            or isinstance(self.column_list()[0],
                          expression.AttributeRef), \
            """assumes first column is the key and second is aggregate result
            column_list: %s""" % self.column_list()

        if self.useKey:
            mapping_var_name = gensym()

            if len(self.grouping_list) == 1:
                produce_template = ct("""%(hashname)s->\
                forall_entries<&%(pipeline_sync)s>\
                ([=](std::pair<const int64_t,int64_t>& %(mapping_var_name)s) {
                    %(output_tuple_type)s %(output_tuple_name)s(\
                    {%(mapping_var_name)s.first, %(mapping_var_name)s.second});
                    %(inner_code)s
                    });
                    """)
            elif len(self.grouping_list) == 2:
                produce_template = ct("""%(hashname)s->\
                forall_entries<&%(pipeline_sync)s>\
                ([=](std::pair<const std::pair<int64_t,int64_t>,int64_t>& \
                %(mapping_var_name)s) {
                    %(output_tuple_type)s %(output_tuple_name)s(\
                    {%(mapping_var_name)s.first.first,\
                    %(mapping_var_name)s.first.second,\
                    %(mapping_var_name)s.second});
                    %(inner_code)s
                    });
                    """)
        else:
            op = self.aggregate_list[0].__class__.__name__
            # translations for Grappa::reduce predefined ops
            coll_op = {'COUNT': 'COLL_ADD',
                       'SUM': 'COLL_ADD',
                       'MAX': 'COLL_MAX',
                       'MIN': 'COLL_MIN'}[op]
            produce_template = ct("""auto %(output_tuple_name)s_tmp = \
            reduce<int64_t, \
            counter, \
            %(coll_op)s, \
            &get_count>\
            (%(hashname)s);

            %(output_tuple_type)s %(output_tuple_name)s;
            %(output_tuple_name)s.set(0, %(output_tuple_name)s_tmp);
            %(inner_code)s
            """)

        pipeline_sync = create_pipeline_synchronization(state)
        get_pipeline_task_name(state)

        # add a dependence on the input aggregation pipeline
        state.addToPipelinePropertySet('dependences', self.input_syncname)

        output_tuple = GrappaStagedTupleRef(gensym(), self.scheme())
        output_tuple_name = output_tuple.name
        output_tuple_type = output_tuple.getTupleTypename()
        state.addDeclarations([output_tuple.generateDefinition()])

        inner_code = self.parent().consume(output_tuple, self, state)
        code = produce_template % locals()
        state.setPipelineProperty("type", "in_memory")
        state.addPipeline(code)

    def consume(self, inputTuple, fromOp, state):
        # save the inter-pipeline task name
        self.input_syncname = get_pipeline_task_name(state)

        inp_sch = self.input.scheme()

        if self.useKey:
            if len(self.grouping_list) == 1:
                materialize_template = ct("""%(hashname)s->update\
                <&%(pipeline_sync)s, int64_t, \
                &Aggregates::%(op)s<int64_t,int64_t>,0>(\
                %(tuple_name)s.get(%(keypos)s),\
                %(tuple_name)s.get(%(valpos)s));
          """)
                # make key from grouped attributes
                keypos = self.grouping_list[0].get_position(inp_sch)

            elif len(self.grouping_list) == 2:
                materialize_template = ct("""%(hashname)s->update\
                <&%(pipeline_sync)s, int64_t, \
                &Aggregates::%(op)s<int64_t,int64_t>,0>(\
                std::pair<int64_t,int64_t>(\
                %(tuple_name)s.get(%(key1pos)s),\
                %(tuple_name)s.get(%(key2pos)s)),\
                %(tuple_name)s.get(%(valpos)s));
          """)
                # make key from grouped attribute
                key1pos = self.grouping_list[0].get_position(inp_sch)
                key2pos = self.grouping_list[1].get_position(inp_sch)
        else:
            # TODO: use optimization for few keys
            # right now it uses key=0
            materialize_template = ct("""%(hashname)s->count = \
            Aggregates::%(op)s<int64_t, int64_t>(%(hashname)s->count, \
                                      %(tuple_name)s.get(%(valpos)s));
            """)

        hashname = self._hashname
        tuple_name = inputTuple.name
        pipeline_sync = state.getPipelineProperty("global_syncname")

        if isinstance(self.aggregate_list[0], expression.ZeroaryOperator):
            # no value needed for Zero-input aggregate,
            # but just provide the first column
            valpos = 0
        elif isinstance(self.aggregate_list[0], expression.UnaryOperator):
            # get value positions from aggregated attributes
            valpos = self.aggregate_list[0].input.get_position(self.scheme())
        else:
            assert False, "only support Unary or Zeroary aggregates"

        op = self.aggregate_list[0].__class__.__name__

        code = materialize_template % locals()
        return code


def wait_statement(name):
    return """{name}.wait();""".format(name=name)


def get_pipeline_task_name(state):
    name = "p_task_{n}".format(n=state.getCurrentPipelineId())
    state.setPipelineProperty('sync', name)
    wait_stmt = wait_statement(name)
    state.addMainWaitStatement(wait_stmt)
    return name


class GrappaHashJoin(algebra.Join, GrappaOperator):
    _i = 0

    @classmethod
    def __genHashName__(cls):
        name = "hash_%03d" % cls._i
        cls._i += 1
        return name

    def produce(self, state):
        if not isinstance(self.condition, expression.EQ):
            msg = "The C compiler can only handle equi-join conditions of\
             a single attribute: %s" % self.condition
            raise ValueError(msg)

        declr_template = ct("""typedef MatchesDHT<int64_t, \
                          %(in_tuple_type)s, std_hash> \
                           DHT_%(in_tuple_type)s;
        DHT_%(in_tuple_type)s %(hashname)s;
        """)

        self.right.childtag = "right"
        self.rightTupleTypeRef = None  # may remain None if CSE succeeds

        left_sch = self.left.scheme()
        right_sch = self.right.scheme()

        # find the attribute that corresponds to the right child
        self.rightCondIsRightAttr = \
            self.condition.right.position >= len(left_sch)
        self.leftCondIsRightAttr = \
            self.condition.left.position >= len(left_sch)
        assert self.rightCondIsRightAttr ^ self.leftCondIsRightAttr, \
            "op: %s,\ncondition: %s, left.scheme: %s, right.scheme: %s" \
            % (self, self.condition, left_sch, right_sch)

        # right key position
        if self.rightCondIsRightAttr:
            self.right_keypos = self.condition.right.position \
                - len(left_sch)
        else:
            self.right_keypos = self.condition.left.position \
                - len(left_sch)

        # left key position
        if self.rightCondIsRightAttr:
            self.left_keypos = self.condition.left.position
        else:
            self.left_keypos = self.condition.right.position

        # common index is defined by same right side and same key
        hashtableInfo = state.lookupExpr((self.right, self.right_keypos))
        if not hashtableInfo:
            # if right child never bound then store hashtable symbol and
            # call right child produce
            self._hashname = self.__genHashName__()
            _LOG.debug("generate hashname %s for %s", self._hashname, self)

            hashname = self._hashname

            # declaration of hash map
            self.rightTupleTypeRef = state.createUnresolvedSymbol()
            in_tuple_type = self.rightTupleTypeRef.getPlaceholder()
            hashdeclr = declr_template % locals()
            state.addDeclarationsUnresolved([hashdeclr])

            init_template = ct("""%(hashname)s.init_global_DHT( &%(hashname)s,
            cores()*16*1024 );""")
            state.addInitializers([init_template % locals()])
            self.right.produce(state)
            state.saveExpr((self.right, self.right_keypos),
                           (self._hashname, self.rightTupleTypename,
                            self.right_syncname))
            # TODO always safe here? I really want to call
            # TODO saveExpr before self.right.produce(),
            # TODO but I need to get the self.rightTupleTypename cleanly
        else:
            # if found a common subexpression on right child then
            # use the same hashtable
            self._hashname, self.rightTupleTypename, self.right_syncname\
                = hashtableInfo
            _LOG.debug("reuse hash %s for %s", self._hashname, self)

        self.left.childtag = "left"
        self.left.produce(state)

    def consume(self, t, src, state):
        if src.childtag == "right":

            right_template = ct("""
            %(hashname)s.insert_async<&%(pipeline_sync)s>(\
            %(keyname)s.get(%(keypos)s), %(keyname)s);
            """)

            hashname = self._hashname
            keyname = t.name

            keypos = self.right_keypos

            self.right_syncname = get_pipeline_task_name(state)

            self.rightTupleTypename = t.getTupleTypename()
            if self.rightTupleTypeRef is not None:
                state.resolveSymbol(self.rightTupleTypeRef,
                                    self.rightTupleTypename)

            pipeline_sync = state.getPipelineProperty('global_syncname')

            # materialization point
            code = right_template % locals()

            return code

        if src.childtag == "left":
            left_template = ct("""
            %(hashname)s.lookup_iter<&%(pipeline_sync)s>( \
            %(keyname)s.get(%(keypos)s), \
            [=](%(right_tuple_type)s& %(right_tuple_name)s) {
              join_coarse_result_count++;
              %(out_tuple_type)s %(out_tuple_name)s = \
               combine<%(out_tuple_type)s, \
                       %(keytype)s, \
                       %(right_tuple_type)s> \
                           (%(keyname)s, %(right_tuple_name)s);
              %(inner_plan_compiled)s
            });
     """)

            # add a dependence on the right pipeline
            state.addToPipelinePropertySet('dependences', self.right_syncname)

            hashname = self._hashname
            keyname = t.name
            keytype = t.getTupleTypename()

            pipeline_sync = state.getPipelineProperty('global_syncname')

            keypos = self.left_keypos

            right_tuple_name = gensym()
            right_tuple_type = self.rightTupleTypename

            outTuple = GrappaStagedTupleRef(gensym(), self.scheme())
            out_tuple_type_def = outTuple.generateDefinition()
            out_tuple_type = outTuple.getTupleTypename()
            out_tuple_name = outTuple.name

            state.addDeclarations([out_tuple_type_def])

            inner_plan_compiled = self.parent().consume(outTuple, self, state)

            code = left_template % locals()
            return code

        assert False, "src not equal to left or right"


def indentby(code, level):
    indent = " " * ((level + 1) * 6)
    return "\n".join([indent + line for line in code.split("\n")])

#
#
#
# class FreeMemory(GrappaOperator):
#  def fire(self, expr):
#    for ref in noReferences(expr)


# Basic selection like serial C++
class GrappaSelect(clangcommon.CSelect, GrappaOperator):
    pass


# Basic apply like serial C++
class GrappaApply(clangcommon.CApply, GrappaOperator):
    pass


# Basic duplication based bag union like serial C++
class GrappaUnionAll(clangcommon.CUnionAll, GrappaOperator):
    pass


# Basic materialized copy based project like serial C++
class GrappaProject(clangcommon.CProject, GrappaOperator):
    pass


class GrappaFileScan(clangcommon.CFileScan, GrappaOperator):
    ascii_scan_template_GRAPH = """
          {
            tuple_graph tg;
            tg = readTuples( "%(name)s" );

            FullEmpty<GlobalAddress<Graph<Vertex>>> f1;
            privateTask( [&f1,tg] {
              f1.writeXF( Graph<Vertex>::create(tg, /*directed=*/true) );
            });
            auto l_%(resultsym)s_index = f1.readFE();

            on_all_cores([=] {
              %(resultsym)s_index = l_%(resultsym)s_index;
            });
        }
        """

    # C++ type inference cannot infer T in readTuples<T>;
    # we resolve it later, so use %%
    ascii_scan_template = """
    {
    if (FLAGS_bin) {
    %(resultsym)s = readTuplesUnordered<%%(result_type)s>( \
    FLAGS_input_file_%(name)s + ".bin" );
    } else {
    %(resultsym)s.data = readTuples<%%(result_type)s>( \
    FLAGS_input_file_%(name)s, FLAGS_nt);
    %(resultsym)s.numtuples = FLAGS_nt;
    auto l_%(resultsym)s = %(resultsym)s;
    on_all_cores([=]{ %(resultsym)s = l_%(resultsym)s; });
    }
    }
    """

    def __get_ascii_scan_template__(self):
        return self.ascii_scan_template

    def __get_binary_scan_template__(self):
        _LOG.warn("binary not currently supported\
         for GrappaLanguage, emitting ascii")
        return self.ascii_scan_template

    def __get_relation_decl_template__(self, name):
        return """
            DEFINE_string(input_file_%(name)s, "%(name)s", "Input file");
            Relation<%(tuple_type)s> %(resultsym)s;
            """


class GrappaStore(clangcommon.BaseCStore, GrappaOperator):
    def __file_code__(self, t, state):
        my_sch = self.scheme()

        filename = (str(self.relation_key).split(":")[2])
        outputnamedecl = """\
        DEFINE_string(output_file, "%s.bin", "Output File");""" % filename
        state.addDeclarations([outputnamedecl])
        names = [x.encode('UTF8') for x in my_sch.get_names()]
        schemefile = 'writeSchema("%s", "%s", "%s");\n' % \
                     (names, my_sch.get_types(), filename)
        state.addPreCode(schemefile)
        resultfile = 'writeTuplesUnordered(&result, "%s.bin");' % filename
        state.addPipelineFlushCode(resultfile)

        return ""


class MemoryScanOfFileScan(rules.Rule):
    """A rewrite rule for making a scan into materialization
     in memory then memory scan"""
    def fire(self, expr):
        if isinstance(expr, algebra.Scan) \
                and not isinstance(expr, GrappaFileScan):
            return GrappaMemoryScan(GrappaFileScan(expr.relation_key,
                                                   expr.scheme()))
        return expr

    def __str__(self):
        return "Scan => MemoryScan(FileScan)"


def grappify(join_type, emit_print):
    return [
        rules.ProjectingJoinToProjectOfJoin(),

        rules.OneToOne(algebra.Select, GrappaSelect),
        MemoryScanOfFileScan(),
        rules.OneToOne(algebra.Apply, GrappaApply),
        rules.OneToOne(algebra.Join, join_type),
        rules.OneToOne(algebra.GroupBy, GrappaGroupBy),
        rules.OneToOne(algebra.Project, GrappaProject),
        rules.OneToOne(algebra.UnionAll, GrappaUnionAll),
        # TODO: obviously breaks semantics
        rules.OneToOne(algebra.Union, GrappaUnionAll),
        clangcommon.StoreToBaseCStore(emit_print, GrappaStore),

        clangcommon.BreakHashJoinConjunction(GrappaSelect, join_type)
    ]


class GrappaAlgebra(Algebra):
    def __init__(self, emit_print=clangcommon.EMIT_CONSOLE):
        self.emit_print = emit_print

    def opt_rules(self, **kwargs):
        # datalog_rules = [
        #     # rules.removeProject(),
        #     rules.CrossProduct2Join(),
        #     rules.SimpleGroupBy(),
        #     # SwapJoinSides(),
        #     rules.OneToOne(algebra.Select, GrappaSelect),
        #     rules.OneToOne(algebra.Apply, GrappaApply),
        #     # rules.OneToOne(algebra.Scan,MemoryScan),
        #     MemoryScanOfFileScan(),
        #     # rules.OneToOne(algebra.Join, GrappaSymmetricHashJoin),
        #     rules.OneToOne(algebra.Join, self.join_type),
        #     rules.OneToOne(algebra.Project, GrappaProject),
        #     rules.OneToOne(algebra.GroupBy, GrappaGroupBy),
        #     # TODO: this Union obviously breaks semantics
        #     rules.OneToOne(algebra.Union, GrappaUnionAll),
        #     rules.OneToOne(algebra.Store, GrappaStore)
        #     # rules.FreeMemory()
        # ]

        join_type = kwargs.get('join_type', GrappaHashJoin)

        # sequence that works for myrial
        rule_grps_sequence = [
            rules.remove_trivial_sequences,
            rules.simple_group_by,
            clangcommon.clang_push_select,
            rules.push_project,
            rules.push_apply,
            grappify(join_type, self.emit_print)
        ]

        if kwargs.get('SwapJoinSides'):
            rule_grps_sequence.insert(0, [rules.SwapJoinSides()])

        return list(itertools.chain(*rule_grps_sequence))



#!/usr/bin/python

import ply.lex as lex

import raco.myrial.exceptions

keywords = ['WHILE', 'DO', 'DEF', 'APPLY', 'CASE', 'WHEN', 'THEN',
            'ELSE', 'END', 'CONST', 'LOAD', 'DUMP', 'UDA', 'TRUE', 'FALSE']

types = ['INT', 'STRING', 'FLOAT', 'BOOLEAN']

comprehension_keywords = ['SELECT', 'AS', 'EMIT', 'FROM', 'WHERE']

word_operators = ['AND', 'OR', 'NOT']

builtins = ['EMPTY', 'WORKER_ID', 'SCAN', 'COUNTALL', 'COUNT', 'STORE',
            'DIFF', 'CROSS', 'JOIN', 'UNIONALL', 'INTERSECT', 'DISTINCT',
            'LIMIT']


# identifiers with special meaning; case-insensitive
reserved = (keywords + types + comprehension_keywords
            + word_operators + builtins)

# Token types; required by ply to have this variable name

tokens = ['LPAREN', 'RPAREN', 'LBRACKET', 'RBRACKET', 'DOT', 'PLUS', 'MINUS',
          'TIMES', 'DIVIDE', 'IDIVIDE', 'LT', 'GT', 'GE', 'LE', 'EQ', 'NE',
          'NE2', 'COMMA', 'SEMI', 'EQUALS', 'COLON', 'DOLLAR', 'ID',
          'STRING_LITERAL', 'INTEGER_LITERAL', 'FLOAT_LITERAL', 'LBRACE',
          'RBRACE'] + reserved

# Regular expression rules for simple tokens
t_LPAREN = r'\('
t_RPAREN = r'\)'
t_LBRACKET = r'\['
t_RBRACKET = r'\]'
t_LBRACE = r'\{'
t_RBRACE = r'\}'

t_PLUS = r'\+'
t_MINUS = r'-'
t_TIMES = r'\*'
t_DIVIDE = r'/'
t_IDIVIDE = r'//'

t_LT = r'<'
t_GT = r'>'
t_LE = r'<='
t_GE = r'>='
t_EQ = r'=='
t_NE = r'!='
t_NE2 = r'<>'

t_DOT = r'\.'
t_COMMA = r','
t_SEMI = r';'
t_EQUALS = r'='
t_COLON = r':'
t_DOLLAR = r'\$'

# Regular expressions for non-trivial tokens


def t_ID(t):
    r'[a-zA-Z_][a-zA-Z_0-9]*'
    global reserved

    upped = t.value.upper()
    if upped in reserved:
        t.type = upped
        t.value = upped
        return t
    else:
        t.type = 'ID'
        return t


def t_FLOAT_LITERAL(t):
    r"""\d*\.\d+"""
    t.value = float(t.value)
    return t


def t_INTEGER_LITERAL(t):
    r'\d+'
    t.value = int(t.value)
    return t


def t_STRING_LITERAL(t):
    r'"([^\\\n"]|\\.)*"'
    t.value = t.value[1:-1].decode("string_escape")
    return t


def t_newline(t):
    r'\n+'
    t.lexer.lineno += len(t.value)


# C-style comments
def t_c_comment(t):
    r'/\*(.|\n)*?\*/'
    t.lexer.lineno += t.value.count('\n')


# database-style comments
def t_db_comment(t):
    r'--.*'


# Always ignore whitespace (spaces and tabs)
t_ignore = ' \t\v'


# Error handling rule
def t_error(t):
    raise raco.myrial.exceptions.MyrialScanException(t)


lexer = lex.lex()




from raco import relation_key
from raco.catalog import Catalog
from raco.algebra import DEFAULT_CARDINALITY
import csv


class ClangTestDatabase(Catalog):
    """
    Interface for table metadata and ingest
    for raco.clang query processor
    """

    def __init__(self):
        # Map from relation keys to tuples of (Bag, scheme.Scheme)
        self.tables = {}

    def get_num_servers(self):
        return 1

    def num_tuples(self, rel_key):
        return DEFAULT_CARDINALITY

    def ingest(self, rel_key, contents, scheme):
        '''Directly load raw data into the database'''
        if isinstance(rel_key, str):
            rel_key = relation_key.RelationKey.from_string(rel_key)
        assert isinstance(rel_key, relation_key.RelationKey)

        with open(rel_key.relation, 'w') as writetable:
            writer = csv.writer(writetable, delimiter=' ')
            for tup in contents:
                writer.writerow(tup)

        self.tables[rel_key] = scheme

    def get_scheme(self, rel_key):
        if isinstance(rel_key, str):
            rel_key = relation_key.RelationKey.from_string(rel_key)

        assert isinstance(rel_key, relation_key.RelationKey)

        scheme = self.tables[rel_key]
        return scheme



import unittest
from raco import RACompiler
import raco.expression as e
import raco.expression.boolean


class ExpressionTest(unittest.TestCase):
    def test_postorder(self):
        expr1 = e.MINUS(e.MAX(e.NamedAttributeRef("salary")),
                        e.MIN(e.NamedAttributeRef("salary")))
        expr2 = e.PLUS(e.LOG(e.NamedAttributeRef("salary")),
                       e.ABS(e.NamedAttributeRef("salary")))

        def isAggregate(expr):
            return isinstance(expr, e.BuiltinAggregateExpression)

        def classname(expr):
            return expr.__class__.__name__

        e1cls = [x for x in expr1.postorder(classname)]

        e2cls = [x for x in expr2.postorder(classname)]

        e1any = any(expr1.postorder(isAggregate))

        e2any = any(expr2.postorder(isAggregate))

        self.assertEqual(str(e1cls), """['NamedAttributeRef', 'MAX', 'NamedAttributeRef', 'MIN', 'MINUS']""")  # noqa
        self.assertEqual(str(e2cls), """['NamedAttributeRef', 'LOG', 'NamedAttributeRef', 'ABS', 'PLUS']""")  # noqa
        self.assertEqual(e1any, True)
        self.assertEqual(e2any, False)

    def test_visitor(self):
        class EvalVisitor(raco.expression.ExpressionVisitor):
            def __init__(self):
                self.stack = []

            def visit_NumericLiteral(self, numericLiteral):
                self.stack.append(numericLiteral.value)

            def visit_NEQ(self, binaryExpr):
                right = self.stack.pop()
                left = self.stack.pop()
                self.stack.append(left != right)

            def visit_AND(self, binaryExpr):
                right = self.stack.pop()
                left = self.stack.pop()
                self.stack.append(left and right)

            def visit_OR(self, binaryExpr):
                right = self.stack.pop()
                left = self.stack.pop()
                self.stack.append(left or right)

            def visit_NOT(self, unaryExpr):
                input = self.stack.pop()
                self.stack.append(not input)

            def visit_GTEQ(self, binaryExpr):
                pass

            def visit_UnnamedAttributeRef(self, unnamed):
                pass

            def visit_StringLiteral(self, stringLiteral):
                pass

            def visit_EQ(self, binaryExpr):
                pass

            def visit_GT(self, binaryExpr):
                pass

            def visit_LT(self, binaryExpr):
                pass

            def visit_LTEQ(self, binaryExpr):
                pass

            def visit_NamedAttributeRef(self, named):
                pass

            def visit_DIVIDE(self, binaryExpr):
                pass

            def visit_IDIVIDE(self, binaryExpr):
                pass

            def visit_MINUS(self, binaryExpr):
                pass

            def visit_NEG(self, binaryExpr):
                pass

            def visit_PLUS(self, binaryExpr):
                pass

            def visit_TIMES(self, binaryExpr):
                pass

        v = EvalVisitor()
        ex = e.AND(e.NEQ(e.NumericLiteral(1), e.NumericLiteral(2)),
                   e.NEQ(e.NumericLiteral(4), e.NumericLiteral(5)))
        ex.accept(v)
        self.assertEqual(v.stack.pop(), True)

        v = EvalVisitor()
        ex = e.AND(e.NEQ(e.NumericLiteral(1), e.NumericLiteral(2)),
                   e.NEQ(e.NumericLiteral(4), e.NumericLiteral(4)))
        ex.accept(v)
        self.assertEqual(v.stack.pop(), False)

        v = EvalVisitor()
        ex = e.AND(e.NEQ(e.NumericLiteral(1), e.NumericLiteral(2)),
                   e.NOT(e.NEQ(e.NumericLiteral(4), e.NumericLiteral(4))))
        ex.accept(v)
        self.assertEqual(v.stack.pop(), True)

        v = EvalVisitor()
        ex = e.NumericLiteral(0xC0FFEE)
        ex.accept(v)
        self.assertEqual(v.stack.pop(), 0xC0FFEE)




from raco.expression import Unbox, UnnamedAttributeRef, NamedAttributeRef
from raco.myrial.exceptions import ColumnIndexOutOfBounds


class EmitArg(object):
    """Defines a single entry in an emit argument list"""

    def expand(self, symbols):
        """Expand an EmitArg into a list of columns.

        symbols is a symbol table that maps relation names to operators
        that define their output.  The order of mappings follows the order
        of arguments in the EMIT clause.

        The output is a list of tuples of the form (name, expression).
        name is a string; expression is an instance of raco.expression.
        """
        raise NotImplementedError()

    def get_statemods(self):
        """Return a list of state variables associated with the emitarg.

        :return: A list of tuples of the form (name, init_expr, update_expr)
        """
        return []


def resolve_attribute_index(idx, symbols):
    """Resolve a column name given a positional index."""

    for op in symbols.values():
        scheme = op.scheme()
        if idx < len(scheme):
            return scheme.getName(idx)
        idx -= len(scheme)

    raise ColumnIndexOutOfBounds(str(idx))


def resolve_unbox(sx, symbols):
    """Resolve a column name given an unbox expression.

    e.g. [FROM A EMIT A.some_column]
    """
    if isinstance(sx.field, basestring):
        return sx.field
    else:
        assert isinstance(sx.field, int)
        op = symbols[sx.relational_expression]
        scheme = op.scheme()
        return scheme.getName(sx.field)


def get_column_name(name, sx, symbols):
    """Create a  column name; generate a name if none was provided.

    :param name: The name supplied by the user, or None if no name provided.
    :param sx: The Expression that defines the output
    :param symbols: A mapping from relation name to Operator instances
    """

    if name:
        return name

    if isinstance(sx, NamedAttributeRef):
        return sx.name
    elif isinstance(sx, UnnamedAttributeRef):
        return resolve_attribute_index(sx.position, symbols)
    elif isinstance(sx, Unbox):
        return resolve_unbox(sx, symbols)
    else:
        return name


class NaryEmitArg(EmitArg):
    """An emit arg that defines one or more columns."""

    def __init__(self, column_names, sexprs, statemods):
        assert column_names is None or len(column_names) == len(sexprs)
        assert len(sexprs) >= 1

        self.column_names = column_names
        self.sexprs = sexprs
        self.statemods = statemods

    def expand(self, symbols):
        names = self.column_names
        if not names:
            names = [None] * len(self.sexprs)

        return [(get_column_name(n, x, symbols), x)
                for n, x in zip(names, self.sexprs)]

    def get_statemods(self):
        return self.statemods


def expand_relation(relation_name, symbols):
    """Expand a given relation into a list of column mappings."""
    assert relation_name in symbols

    op = symbols[relation_name]
    scheme = op.scheme()

    colnames = [x[0] for x in iter(scheme)]
    return [(colname, Unbox(relation_name, colname))
            for colname in colnames]


class TableWildcardEmitArg(EmitArg):
    """An emit arg that refers to all columns in a table.

    e.g., [FROM emp, dept EMIT dept.*]"""

    def __init__(self, relation_name):
        self.relation_name = relation_name

    def expand(self, symbols):
        return expand_relation(self.relation_name, symbols)


class FullWildcardEmitArg(EmitArg):
    """Emit all columns from the input.

    This is basically 'select *'
    """

    def __init__(self):
        pass

    def expand(self, symbols):
        cols = []
        for relation_name in symbols:
            cols.extend(expand_relation(relation_name, symbols))
        return cols



"""
A parser for Purple programs.

The result is a parse object that can return a (recursive) relational algebra
expression.
"""

from pyparsing import Literal, CaselessLiteral, Word, delimitedList, \
    Optional, Combine, Group, alphas, nums, alphanums, oneOf, quotedString, \
    ZeroOrMore, restOfLine

import raco
from raco import expression
import raco.datalog.model as model


def show(x):
    print x
    return x

drop = lambda x: Literal(x).suppress()

# define Datalog tokens
ident = Word(alphas, alphanums + "_$")
predicate = ident.setName("Predicate")

E = CaselessLiteral("E")

# Get all the aggregate expression classes
aggregate_functions = raco.expression.aggregate_functions()

# All binary operators
binopclasses = expression.binary_ops()

# a list of string literals representing opcodes
opcodes = sum([oc.literals for oc in binopclasses], [])

binopstr = " ".join(opcodes)


def parsebinop(opexpr):
    "parse action for binary operators"
    left, opstr, right = opexpr

    for opclass in binopclasses:
        if opstr in opclass.literals:
            return opclass(left, right)

binop = oneOf(binopstr)
arithSign = Word("+-", exact=1)

realNum = Combine(Optional(arithSign) +
                  (Word(nums) + "." + Optional(Word(nums)) | ("." + Word(nums)))  # noqa
                  + Optional(E + Optional(arithSign) + Word(nums)))
realNum.setParseAction(lambda x: expression.NumericLiteral(float(x[0])))

intNum = Combine(Optional(arithSign) + Word(nums) +
                 Optional(E + Optional("+") + Word(nums)))
intNum.setParseAction(lambda x: expression.NumericLiteral(int(x[0])))

number = realNum | intNum

variable = ident.copy()
variable.setParseAction(lambda x: model.Var(x[0]))

quotedString.setParseAction(lambda x: expression.StringLiteral(x[0][1:-1]))

literal = quotedString | number

valueref = variable | literal


def mkterm(x):
    return model.Term(x)

term = (predicate
        + drop("(")
        + Group(delimitedList(valueref, ","))
        + drop(")")).setParseAction(mkterm)


def checkval(xs):
    left, op, right = xs[0]
    if op == '=':
        result = left == right
    else:
        result = eval(left + op + right)
    return result

groundcondition = Group(literal + binop + literal)
# groundcondition.setParseAction(checkval)

# TODO: deeper expression trees
condition = (valueref + binop + valueref)
condition.setParseAction(parsebinop)

body = delimitedList(term | groundcondition | condition, ",")
# .setParseAction(show) # lambda xs: [Term(x) for x in xs])

partitioner = drop("h(") + delimitedList(variable, ",") + drop(")")
partitioner.setParseAction(lambda x: model.PartitionBy(x))

allservers = Literal("*").setParseAction(lambda x: model.Broadcast())
server = drop("@") + (partitioner | allservers)
timeexpr = variable + oneOf("+ -") + Word(nums)
timeexpr.setParseAction(lambda xs: "".join([str(x) for x in xs]))

timestep = drop("#") + (intNum | timeexpr | variable)
timestep.setParseAction(lambda x: model.Timestep(x[0]))

# expressions without aggregates
# TODO more complete
simpleArithExpression = (valueref + binop + valueref)
simpleArithExpression.setParseAction(parsebinop)


def mkagg(x):
    opstr, arg = x
    for aggclass in aggregate_functions:
        if opstr.lower() == aggclass.__name__.lower():
            return aggclass(arg)
    raise "Aggregate Function %s not found among %s" % (opstr, aggregate_functions)  # noqa

aggregate = (Word(alphas) + drop("(") + variable + drop(")")) | \
            (Word(alphas) + drop("(") + simpleArithExpression + drop(")"))
aggregate.setParseAction(mkagg)


# expressions containing aggregates
# TODO deeper instead of enumeration
arithExpression = (aggregate + binop + aggregate) | \
                  (valueref + binop + valueref) | \
                  (aggregate + binop + valueref) | \
                  (valueref + binop + aggregate)
arithExpression.setParseAction(parsebinop)

# greedy parsing so put arithExpression first
headvalueref = arithExpression | aggregate | variable | literal

headterm = (predicate + Optional(server)
            + drop("(") + Group(delimitedList(headvalueref, ",")) + drop(")"))


def mkIDB(x):
    if len(x) == 4:
        idb = model.IDB(mkterm((x[0], x[2])), x[1], x[3])
    elif len(x) == 3:
        if isinstance(x[2], model.Timestep):
            idb = model.IDB(mkterm((x[0], x[1])), timestep=x[2])
        else:
            idb = model.IDB(mkterm((x[0], x[2])), x[1])
    else:
        idb = model.IDB(mkterm(x))
    return idb

head = (headterm + Optional(timestep) + drop(":-")).setParseAction(mkIDB)
# head.setParseAction(show)


def mkrule(x):
    """Workaround for AttributeError: Class Rule has no __call__ method when
    running through wsgi"""
    return model.Rule(x)

rule = (head + Group(body)
        + Optional(drop(";")) + Optional(drop(".")))
rule.setParseAction(mkrule)


def mkprogram(x):
    """Workaround for AttributeError: Class Rule has no __call__ method when
    running through wsgi"""
    return model.Program(x)

comment = (Literal("#") + restOfLine).suppress()

program = ZeroOrMore(rule | comment).setParseAction(mkprogram)


def parse(query):
    return program.parseString(query)[0]



"""
Boolean operators for use in Raco expression trees
"""

from .expression import Expression, UnaryOperator, BinaryOperator, \
    AttributeRef, NumericLiteral, check_type, TypeSafetyViolation

import raco.types

import abc


class BooleanExpression(Expression):
    pass


class UnaryBooleanOperator(BooleanExpression, UnaryOperator):
    def typeof(self, scheme, state_scheme):
        lt = self.input.typeof(scheme, state_scheme)
        check_type(lt, raco.types.BOOLEAN_TYPE)
        return raco.types.BOOLEAN_TYPE


class BinaryBooleanOperator(BooleanExpression, BinaryOperator):
    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        check_type(lt, raco.types.BOOLEAN_TYPE)
        rt = self.left.typeof(scheme, state_scheme)
        check_type(rt, raco.types.BOOLEAN_TYPE)
        return raco.types.BOOLEAN_TYPE


class BinaryComparisonOperator(BinaryBooleanOperator):
    def typeof(self, scheme, state_scheme):
        lt = self.left.typeof(scheme, state_scheme)
        rt = self.right.typeof(scheme, state_scheme)
        if lt == rt:
            return raco.types.BOOLEAN_TYPE
        if lt in raco.types.NUMERIC_TYPES and rt in raco.types.NUMERIC_TYPES:
            return raco.types.BOOLEAN_TYPE
        else:
            raise TypeSafetyViolation("Can't compare %s and %s", (lt, rt))


class NOT(UnaryBooleanOperator):
    literals = ["not", "NOT", "-"]

    def evaluate(self, _tuple, scheme, state=None):
        return not self.input.evaluate(_tuple, scheme, state)


class AND(BinaryBooleanOperator):
    literals = ["and", "AND"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) and
                self.right.evaluate(_tuple, scheme, state))


class OR(BinaryBooleanOperator):
    literals = ["or", "OR"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) or
                self.right.evaluate(_tuple, scheme, state))


class EQ(BinaryComparisonOperator):
    literals = ["=", "=="]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) ==
                self.right.evaluate(_tuple, scheme, state))


class LT(BinaryComparisonOperator):
    literals = ["<", "lt"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) <
                self.right.evaluate(_tuple, scheme, state))


class GT(BinaryComparisonOperator):
    literals = [">", "gt"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) >
                self.right.evaluate(_tuple, scheme, state))


class GTEQ(BinaryComparisonOperator):
    literals = [">=", "gteq", "gte"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) >=
                self.right.evaluate(_tuple, scheme, state))


class LTEQ(BinaryComparisonOperator):
    literals = ["<=", "lteq", "lte"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) <=
                self.right.evaluate(_tuple, scheme, state))


class NEQ(BinaryComparisonOperator):
    literals = ["!=", "neq", "ne"]

    def evaluate(self, _tuple, scheme, state=None):
        return (self.left.evaluate(_tuple, scheme, state) !=
                self.right.evaluate(_tuple, scheme, state))

reverse = {
    NEQ: NEQ,
    EQ: EQ,
    GTEQ: LTEQ,
    LTEQ: GTEQ,
    GT: LT,
    LT: GT
}

TAUTOLOGY = EQ(NumericLiteral(1), NumericLiteral(1))


def extract_conjuncs(sexpr):
    """Return a list of conjunctions from a scalar expression."""

    if isinstance(sexpr, AND):
        left = extract_conjuncs(sexpr.left)
        right = extract_conjuncs(sexpr.right)
        return left + right
    else:
        return [sexpr]

from .util import toUnnamed
